@@ -15,8 +15,11 @@ use constant {
 	# Lowest compat level that does *not* cause deprecation
 	# warnings
 	'LOWEST_NON_DEPRECATED_COMPAT_LEVEL' => 9,
+	# Lowest compat level to generate "debhelper-compat (= X)"
+	# relations for.
+	'LOWEST_VIRTUAL_DEBHELPER_COMPAT_LEVEL' => 9,
 	# Highest compat level permitted
-	'MAX_COMPAT_LEVEL' => 12,
+	'MAX_COMPAT_LEVEL' => 13,
 	# Magic value for xargs
 	'XARGS_INSERT_PARAMS_HERE' => \'<INSERT-HERE>', #'# Hi emacs.
 	# Magic value for debhelper tools to request "current version"
@@ -111,7 +114,7 @@ qw(
 	glob_expand_error_handler_reject
 	glob_expand_error_handler_warn_and_discard
 	glob_expand_error_handler_silently_ignore
-
+	glob_expand_error_handler_reject_nomagic_warn_discard
 ),
 	# Generate triggers, substvars, maintscripts, build-time temporary files
 qw(
@@ -143,6 +146,7 @@ qw(
 	is_cross_compiling
 	is_build_profile_active
 	get_buildoption
+	perl_cross_incdir
 ),
 	# Other
 qw(
@@ -163,6 +167,7 @@ qw(
 	DEFAULT_PACKAGE_TYPE
 	DBGSYM_PACKAGE_TYPE
 	DH_BUILTIN_VERSION
+	is_known_package
 	assert_opt_is_known_package
 	restore_all_files
 
@@ -181,10 +186,35 @@ our $PKGVERSION_REGEX = qr/
                  [0-9][0-9A-Za-z.+:~]*       # Upstream version (with no hyphens)
                  (?: - [0-9A-Za-z.+:~]+ )*   # Optional debian revision (+ upstreams versions with hyphens)
                           /xoa;
+our $MAINTSCRIPT_TOKEN_REGEX = qr/[A-Za-z0-9_.+]+/o;
+
+# From Policy 5.1:
+#
+#  The field name is composed of US-ASCII characters excluding control
+#  characters, space, and colon (i.e., characters in the ranges U+0021
+#  (!) through U+0039 (9), and U+003B (;) through U+007E (~),
+#  inclusive). Field names must not begin with the comment character
+#  (U+0023 #), nor with the hyphen character (U+002D -).
+our $DEB822_FIELD_REGEX = qr/
+	    [\x21\x22\x24-\x2C\x2F-\x39\x3B-\x7F]  # First character
+	    [\x21-\x39\x3B-\x7F]*                  # Subsequent characters (if any)
+    /xoa;
+
+our $PARSE_DH_SEQUENCE_INFO = 0;
+
+# We need logging in compat 9 or in override/hook targets (for --remaining-packages to work)
+# - This option is a global toggle to disable logs for special commands (e.g. dh or dh_clean)
+# It is initialized during "init".  This implies that commands that never calls init are
+# not dh_* commands or do not need the log
+my $write_log = undef;
 
 sub init {
 	my %params=@_;
 
+	if ($params{internal_parse_dh_sequence_info}) {
+		$PARSE_DH_SEQUENCE_INFO = 1;
+	}
+
 	# Check if we can by-pass the expensive Getopt::Long by optimising for the
 	# common case of "-a" or "-i"
 	if (scalar(@ARGV) == 1 && ($ARGV[0] eq '-a' || $ARGV[0] eq '-i') &&
@@ -243,7 +273,7 @@ sub init {
 	# make sure verbose is on. Otherwise, check DH_QUIET.
 	if (defined $ENV{DH_VERBOSE} && $ENV{DH_VERBOSE} ne "") {
 		$dh{VERBOSE}=1;
-	} elsif (defined $ENV{DH_QUIET} && $ENV{DH_QUIET} ne "") {
+	} elsif (defined $ENV{DH_QUIET} && $ENV{DH_QUIET} ne "" || get_buildoption("terse")) {
 		$dh{QUIET}=1;
 	}
 
@@ -285,17 +315,25 @@ sub init {
 	}
 
 	$dh{U_PARAMS} //= [];
+
+	if ($params{'inhibit_log'}) {
+		$write_log = 0;
+	} else {
+		# Only initialize if unset (i.e. avoid overriding an early call
+		# to inhibit_log()
+		$write_log //= 1;
+	}
 }
 
-# Run at exit. Add the command to the log files for the packages it acted
-# on, if it's exiting successfully.
-my $write_log=1;
+# Ensure the log is written if requested but only if the command was
+# successful.
 sub END {
+	return if $? != 0 or not $write_log;
 	# If there is no 'debian/control', then we are not being run from
 	# a package directory and then the write_log will not do what we
 	# expect.
 	return if not -f 'debian/control';
-	if ($? == 0 && $write_log && (compat(9, 1) || $ENV{DH_INTERNAL_OVERRIDE})) {
+	if (compat(9, 1) || $ENV{DH_INTERNAL_OVERRIDE}) {
 		write_log(basename($0), @{$dh{DOPACKAGES}});
 	}
 }
@@ -306,20 +344,6 @@ sub logfile {
 	return "debian/${ext}debhelper.log"
 }
 
-sub add_override {
-	my $line=shift;
-	$line="override_$ENV{DH_INTERNAL_OVERRIDE} $line"
-		if defined $ENV{DH_INTERNAL_OVERRIDE};
-	return $line;
-}
-
-sub remove_override {
-	my $line=shift;
-	$line=~s/^\Qoverride_$ENV{DH_INTERNAL_OVERRIDE}\E\s+//
-		if defined $ENV{DH_INTERNAL_OVERRIDE};
-	return $line;
-}
-
 sub load_log {
 	my ($package, $db)=@_;
 
@@ -327,7 +351,7 @@ sub load_log {
 	open(LOG, "<", logfile($package)) || return;
 	while (<LOG>) {
 		chomp;
-		my $command=remove_override($_);
+		my $command = $_;
 		push @log, $command;
 		$db->{$package}{$command}=1 if defined $db;
 	}
@@ -342,9 +366,9 @@ sub write_log {
 	return if $dh{NO_ACT};
 
 	foreach my $package (@packages) {
-		my $log=logfile($package);
+		my $log = logfile($package);
 		open(LOG, ">>", $log) || error("failed to write to ${log}: $!");
-		print LOG add_override($cmd)."\n";
+		print LOG $cmd."\n";
 		close LOG;
 	}
 }
@@ -355,8 +379,8 @@ sub commit_override_log {
 	return if $dh{NO_ACT};
 
 	foreach my $package (@packages) {
-		my @log=map { remove_override($_) } load_log($package);
-		my $log=logfile($package);
+		my @log = load_log($package);
+		my $log = logfile($package);
 		open(LOG, ">", $log) || error("failed to write to ${log}: $!");
 		print LOG $_."\n" foreach @log;
 		close LOG;
@@ -438,9 +462,10 @@ sub _doit {
 		if (defined($options)) {
 			if (defined(my $dir = $options->{chdir})) {
 				if ($dir ne '.') {
-					chdir($dir) or error("chdir(\"${dir}\) failed: $!");
+					chdir($dir) or error("chdir(\"${dir}\") failed: $!");
 				}
 			}
+			open(STDIN, '<', '/dev/null') or error("redirect STDIN failed: $!");
 			if (defined(my $output = $options->{stdout})) {
 				open(STDOUT, '>', $output) or error("redirect STDOUT failed: $!");
 			}
@@ -742,18 +767,57 @@ sub nonquiet_print {
 	}
 }
 
-# Output an error message and die (can be caught).
-sub error {
-	my $message=shift;
+{
+	my $_use_color;
+	sub _color {
+		my ($msg, $color) = @_;
+		if (not defined($_use_color)) {
+			# This part is basically Dpkg::ErrorHandling::setup_color over again
+			# with some tweaks.
+			# (but the module uses Dpkg + Dpkg::Gettext, so it is very expensive
+			# to load)
+			my $mode = $ENV{'DH_COLORS'} // $ENV{'DPKG_COLORS'};
+			# Support NO_COLOR (https://no-color.org/)
+			$mode //= exists($ENV{'NO_COLOR'}) ? 'never' : 'auto';
 
-	die basename($0).": $message\n";
-}
+			if ($mode eq 'auto') {
+				$_use_color = 1 if -t *STDOUT or -t *STDERR;
+			} elsif ($mode eq 'always') {
+				$_use_color = 1;
+			} else {
+				$_use_color = 0;
+			}
 
-# Output a warning.
-sub warning {
-	my $message=shift;
-	
-	print STDERR basename($0).": $message\n";
+			eval {
+				require Term::ANSIColor if $_use_color;
+			};
+			if ($@) {
+				# In case of errors, skip colors.
+				$_use_color = 0;
+			}
+		}
+		if ($_use_color) {
+			local $ENV{'NO_COLOR'} = undef;
+			$msg = Term::ANSIColor::colored($msg, $color);
+		}
+		return $msg;
+	}
+
+	# Output an error message and die (can be caught).
+	sub error {
+		my ($message) = @_;
+		# ensure the error code is well defined.
+		$! = 255;
+		die(_color(basename($0), 'bold') . ': ' . _color('error', 'bold red') . ": $message\n");
+	}
+
+	# Output a warning.
+	sub warning {
+		my ($message) = @_;
+		$message //= '';
+
+		print STDERR _color(basename($0), 'bold') . ': ' . _color('warning', 'bold yellow') . ": $message\n";
+	}
 }
 
 # Returns the basename of the argument passed to it.
@@ -815,6 +879,9 @@ my $compat_from_bd;
 						warning("Please specify the debhelper compat level exactly once.");
 						warning(" * debian/compat requests compat ${new_compat}.");
 						warning(" * debian/control requests compat ${compat_from_bd} via \"debhelper-compat (= ${compat_from_bd})\"");
+						warning();
+						warning("Hint: If you just added a build-dependency on debhelper-compat, then please remember to remove debian/compat");
+						warning();
 						error("debhelper compat level specified both in debian/compat and via build-dependency on debhelper-compat");
 					}
 					$c = $new_compat;
@@ -948,11 +1015,7 @@ sub default_sourcedir {
 			}
 		}
 		foreach my $file (@try) {
-			if (-f $file &&
-				(! $dh{IGNORE} || ! exists $dh{IGNORE}->{$file})) {
-				return $file;
-			}
-
+			return $file if -f $file;
 		}
 
 		return "";
@@ -988,22 +1051,33 @@ sub pkgfilename {
 # As a side effect, sets $dh{VERSION} to the version of this package.
 {
 	# Caches return code so it only needs to run dpkg-parsechangelog once.
-	my %isnative_cache;
+	my (%isnative_cache, %pkg_version);
 	
 	sub isnative {
-		my $package=shift;
+		my ($package) = @_;
+		my $cache_key = $package;
 
-		return $isnative_cache{$package} if defined $isnative_cache{$package};
+		if (exists($isnative_cache{$cache_key})) {
+			$dh{VERSION} = $pkg_version{$cache_key};
+			return $isnative_cache{$cache_key};
+		}
+
+		# Make sure we look at the correct changelog.
+		my $isnative_changelog = pkgfile($package,"changelog");
+		if (! $isnative_changelog) {
+			$isnative_changelog = "debian/changelog";
+			$cache_key = '_source';
+			# check if we looked up the default changelog
+			if (exists($isnative_cache{$cache_key})) {
+				$dh{VERSION} = $pkg_version{$cache_key};
+				return $isnative_cache{$cache_key};
+			}
+		}
 
 		if (not %isnative_cache) {
 			require Dpkg::Changelog::Parse;
 		}
 
-		# Make sure we look at the correct changelog.
-		my $isnative_changelog=pkgfile($package,"changelog");
-		if (! $isnative_changelog) {
-			$isnative_changelog="debian/changelog";
-		}
 		my $res = Dpkg::Changelog::Parse::changelog_parse(
 			file => $isnative_changelog,
 			compression => 0,
@@ -1016,15 +1090,15 @@ sub pkgfilename {
 		if (not defined($version) or not $version->is_valid) {
 			error("changelog parse failure; invalid or missing version");
 		}
-		# Get the package version.
-		$dh{VERSION} = $version->as_string;
+		# Get and cache the package version.
+		$dh{VERSION} = $pkg_version{$cache_key} = $version->as_string;
 
 		# Is this a native Debian package?
 		if (index($dh{VERSION}, '-') > -1) {
-			return $isnative_cache{$package}=0;
+			return $isnative_cache{$cache_key} = 0;
 		}
 		else {
-			return $isnative_cache{$package}=1;
+			return $isnative_cache{$cache_key} = 1;
 		}
 	}
 }
@@ -1130,7 +1204,7 @@ sub autoscript_sed {
 	my ($sed, $infile, $outfile, $out_fd) = @_;
 	if (not defined($sed) or ref($sed)) {
 		my $out = $out_fd;
-		open(my $in, '<', $infile) or die "$infile: $!";
+		open(my $in, '<', $infile) or error("open $infile failed: $!");
 		if (not defined($out_fd)) {
 			open($out, '>>', $outfile) or error("open($outfile): $!");
 		}
@@ -1145,9 +1219,9 @@ sub autoscript_sed {
 			}
 		}
 		if (not defined($out_fd)) {
-			close($out) or error("close($outfile): $!");
+			close($out) or error("close $outfile failed: $!");
 		}
-		close($in) or die "$infile: $!";
+		close($in) or error("close $infile failed: $!");
 	}
 	else {
 		error("Internal error - passed open handle for legacy method") if defined($out_fd);
@@ -1292,6 +1366,20 @@ sub glob_expand_error_handler_warn_and_discard {
 	return;
 }
 
+# Emulates the "old" glob mechanism; not recommended for new code as
+# it permits some globs expand to nothing with only a warning.
+sub glob_expand_error_handler_reject_nomagic_warn_discard {
+	my ($pattern, $dir_ref) = @_;
+	for my $dir (@{$dir_ref}) {
+		my $full_pattern = "$dir/$pattern";
+		my @matches = bsd_glob($full_pattern, GLOB_CSH & ~(GLOB_TILDE));
+		if (@matches) {
+			goto \&glob_expand_error_handler_reject;
+		}
+	}
+	goto \&glob_expand_error_handler_warn_and_discard;
+}
+
 sub glob_expand_error_handler_silently_ignore {
 	return;
 }
@@ -1321,38 +1409,116 @@ sub glob_expand {
 	return @result;
 }
 
+
+my %BUILT_IN_SUBST = (
+	'Space'        => ' ',
+	'Dollar'       => '$',
+	'Newline'      => "\n",
+	'Tab'          => "\b",
+);
+
+sub _variable_substitution {
+	my ($text, $loc) = @_;
+	return $text if index($text, '$') < 0;
+	my $pos = -1;
+	my $subst_count = 0;
+	my $expansion_count = 0;
+	my $current_size = length($text);
+	my $expansion_size_limit = 3 * $current_size;
+	1 while ($text =~ s<
+			\$\{([A-Za-z0-9][-_:0-9A-Za-z]*)\}  # Match ${something} and replace it
+		>[
+			my $match = $1;
+			my $new_pos = pos()//-1;
+			my $value;
+
+			if ($pos == $new_pos) {
+				# Safe-guard in case we ever implement recursive expansion
+				error("Error substituting in ${loc} (at position $pos); recursion limit while expanding \${${match}")
+					if (++$subst_count >= 20);
+			} else {
+				$subst_count = 0;
+				$pos = $new_pos;
+				if (++$expansion_count >= 50) {
+					error("Error substituting in ${loc}; substitution limit of ${expansion_count} reached");
+				}
+			}
+			if (exists($BUILT_IN_SUBST{$match})) {
+				$value = $BUILT_IN_SUBST{$match};
+			} elsif ($match =~ m/^DEB_(?:BUILD|HOST|TARGET)_/) {
+				$value = dpkg_architecture_value($match) //
+					error(qq{Cannot expand "\${${match}}\" in ${loc} as it is not a known dpkg-architecture value});
+			} elsif ($match =~ m/^env:(.+)/) {
+				my $env_var = $1;
+				$value = $ENV{$env_var} //
+					error(qq{Cannot expand "\${${match}}" in ${loc} as the ENV variable "${env_var}" is unset});
+			}
+			error("Cannot resolve variable \${$match} in ${loc}")
+				if not defined($value);
+			# We do not support recursive expansion.
+			$value =~ s/\$/\$\{\}/;
+			$current_size += length($value) - length($match) - 3;
+			if ($current_size > 4096 and $current_size > $expansion_size_limit) {
+				error("Refusing to expand \${${match}} in ${loc} - the original input seems to grow beyond reasonable'
+						 . ' limits!");
+			}
+			$value;
+		]gex);
+	$text =~ s/\$\{\}/\$/g;
+
+	return $text;
+}
+
 # Reads in the specified file, one line at a time. splits on words, 
 # and returns an array of arrays of the contents.
 # If a value is passed in as the second parameter, then glob
 # expansion is done in the directory specified by the parameter ("." is
 # frequently a good choice).
+# In compat 13+, it will do variable expansion (after splitting the lines
+# into words)
 sub filedoublearray {
 	my ($file, $globdir, $error_handler) = @_;
 
 	# executable config files are a v9 thing.
 	my $x=! compat(8) && -x $file;
+	my $expand_patterns = compat(12) ? 0 : 1;
+	my $source;
 	if ($x) {
 		require Cwd;
 		my $cmd=Cwd::abs_path($file);
 		$ENV{"DH_CONFIG_ACT_ON_PACKAGES"} = join(",", @{$dh{"DOPACKAGES"}});
-		open (DH_FARRAY_IN, "$cmd |") || error("cannot run $file: $!");
+		open(DH_FARRAY_IN, '-|', $cmd) || error("cannot run $file: $!");
 		delete $ENV{"DH_CONFIG_ACT_ON_PACKAGES"};
+		$source = "output of ./${file}";
 	}
 	else {
 		open (DH_FARRAY_IN, '<', $file) || error("cannot read $file: $!");
+		$source = $file;
 	}
 
 	my @ret;
 	while (<DH_FARRAY_IN>) {
 		chomp;
-		if (not $x)  {
-			next if /^#/ || /^$/;
+		if ($x) {
+			if (m/^\s++$/) {
+				error("Executable config file $file produced a non-empty whitespace-only line");
+			}
+		} else {
+			s/^\s++//;
+			next if /^#/;
+			s/\s++$//;
 		}
+		# We always ignore/permit empty lines
+		next if $_ eq '';
 		my @line;
+		my $source_ref = "${source} (line $.)";
 
 		if (defined($globdir) && ! $x) {
 			if (ref($globdir)) {
 				my @patterns = split;
+				if ($expand_patterns) {
+					@patterns = map {_variable_substitution($_, $source_ref)} @patterns;
+				}
 				push(@line, glob_expand($globdir, $error_handler, @patterns));
 			} else {
 				# Legacy call - Silently discards globs that match nothing.
@@ -1362,19 +1528,40 @@ sub filedoublearray {
 				# filenames that come out are relative to it.
 				foreach (map { glob "$globdir/$_" } split) {
 					s#^$globdir/##;
+					if ($expand_patterns) {
+						$_ = _variable_substitution($_, $source_ref);
+					}
 					push @line, $_;
 				}
 			}
 		}
 		else {
 			@line = split;
+			if ($expand_patterns) {
+				@line = map {_variable_substitution($_, $source_ref)} @line;
+			}
 		}
 		push @ret, [@line];
 	}
 
 	if (!close(DH_FARRAY_IN)) {
 		if ($x) {
-			error("Error closing fd/process for $file: $!") if $!;
+			my ($err, $proc_err) = ($!, $?);
+			error("Error closing fd/process for $file: $err") if $err;
+			# The interpreter did not like the file for some reason.
+			# Lets check if the maintainer intended it to be
+			# executable.
+			if (not is_so_or_exec_elf_file($file) and not _has_shbang_line($file)) {
+				warning("$file is marked executable but does not appear to an executable config.");
+				warning();
+				warning("If $file is intended to be an executable config file, please ensure it can");
+				warning("be run as a stand-alone script/program (e.g. \"./${file}\")");
+				warning("Otherwise, please remove the executable bit from the file (e.g. chmod -x \"${file}\")");
+				warning();
+				warning('Please see "Executable debhelper config files" in debhelper(7) for more information.');
+				warning();
+			}
+			$? = $proc_err;
 			error_exitcode("$file (executable config)");
 		} else {
 			error("problem reading $file: $!");
@@ -1482,13 +1669,14 @@ sub is_cross_compiling {
 # As a side effect, populates %package_arches and %package_types
 # with the types of all packages (not only those returned).
 my (%package_types, %package_arches, %package_multiarches, %packages_by_type,
-    %package_sections, $sourcepackage, %package_cross_type);
+    %package_sections, $sourcepackage, %package_cross_type, %dh_bd_sequences);
 
 # Resets the arrays; used mostly for testing
 sub resetpackages {
 	undef $sourcepackage;
 	%package_types = %package_arches = %package_multiarches =
 	    %packages_by_type = %package_sections = %package_cross_type = ();
+	%dh_bd_sequences = ();
 }
 
 # Returns source package name
@@ -1504,17 +1692,20 @@ sub getpackages {
 
 	$type //= 'all-listed-in-control-file';
 
-	if (%packages_by_type) {
-		return @{$packages_by_type{$type}};
+	if (not %packages_by_type) {
+		_parse_debian_control();
 	}
+	return @{$packages_by_type{$type}};
+}
 
-	my $fd;
+sub _parse_debian_control {
 	my $package="";
 	my $arch="";
 	my $section="";
+	my $valid_pkg_re = qr{^${PKGNAME_REGEX}$}o;
 	my ($package_type, $multiarch, %seen, @profiles, $source_section,
 		$included_in_build_profile, $cross_type, $cross_target_arch,
-		%bd_fields, $bd_field_value);
+		%bd_fields, $bd_field_value, %seen_fields, $fd);
 	if (exists $ENV{'DEB_BUILD_PROFILES'}) {
 		@profiles=split /\s+/, $ENV{'DEB_BUILD_PROFILES'};
 	}
@@ -1530,29 +1721,56 @@ sub getpackages {
 		s/\s+$//;
 		next if m/^\s*+\#/;
 
-		if (/^Source:\s*+(.*)/i) {
-			$sourcepackage = $1;
-			$bd_field_value = undef;
+		if (/^\s/) {
+			if (not %seen_fields) {
+				error("Continuation line seen before first stanza in debian/control (line $.)");
+			}
+			# Continuation line
+			push(@{$bd_field_value}, $_) if $bd_field_value;
+		} elsif (not $_ and not %seen_fields) {
+			# Ignore empty lines before first stanza
 			next;
-		} elsif (/^Section:\s*+(.*)$/i) {
-			$source_section = $1;
-			$bd_field_value = undef;
-			next;
-		} elsif (/^(Build-Depends(?:-Arch|-Indep)?):\s*+(.*)$/i) {
-			my ($field, $value) = (lc($1), $2);
-			$bd_field_value = [$value];
-			$bd_fields{$field} = $bd_field_value;
-		} elsif (/^\S/) {
-			$bd_field_value = undef;
-		} elsif (/^\s/ and $bd_field_value) {
-			push(@{$bd_field_value}, $_);
+		} elsif ($_) {
+			my ($field_name, $value);
+
+			if (m/^($DEB822_FIELD_REGEX):\s*(.*)/o) {
+				($field_name, $value) = (lc($1), $2);
+				if (exists($seen_fields{$field_name})) {
+					my $first_time = $seen_fields{$field_name};
+					error("${field_name}-field appears twice in the same stanza of debian/control. " .
+						  "First time on line $first_time, second time: $.");
+				}
+				$seen_fields{$field_name} = $.;
+				$bd_field_value = undef;
+			} else {
+				# Invalid file
+				error("Parse error in debian/control, line $., read: $_");
+			}
+			if ($field_name eq 'source') {
+				$sourcepackage = $value;
+				if ($sourcepackage !~ $valid_pkg_re) {
+					error('Source-field must be a valid package name, ' .
+						  "got: \"${sourcepackage}\", should match \"${valid_pkg_re}\"");
+				}
+				next;
+			} elsif ($field_name eq 'section') {
+				$source_section = $value;
+				next;
+			} elsif ($field_name =~ /^(?:build-depends(?:-arch|-indep)?)$/) {
+				$bd_field_value = [$value];
+				$bd_fields{$field_name} = $bd_field_value;
+			}
 		}
-		next if not $_ and not defined($sourcepackage);
-		last if (!$_ or eof); # end of stanza.
+		last if not $_ or eof;
 	}
 	error("could not find Source: line in control file.") if not defined($sourcepackage);
 	if (%bd_fields) {
 		my ($dh_compat_bd, $final_level);
+		my %field2addon_type = (
+			'build-depends' => 'both',
+			'build-depends-arch' => 'arch',
+			'build-depends-indep' => 'indep',
+		);
 		for my $field (sort(keys(%bd_fields))) {
 			my $value = join(' ', @{$bd_fields{$field}});
 			$value =~ s/^\s*//;
@@ -1591,6 +1809,28 @@ sub getpackages {
 					warning("   compat level into the file debian/compat.  (E.g. \"echo ${clevel} > debian/compat\")");
 					error("Could not parse desired debhelper compat level from relation: $dep");
 				}
+				# Build-Depends on dh-sequence-<foo> OR dh-sequence-<foo> (<op> <version>)
+				if ($PARSE_DH_SEQUENCE_INFO and $dep =~ m/^dh-sequence-(${PKGNAME_REGEX})\s*(?:[(]\s*(?:[<>]?=|<<|>>)\s*(?:${PKGVERSION_REGEX})\s*[)])?(\s*[^\|]+[]>]\s*)?$/) {
+					my $sequence = $1;
+					my $has_profile_or_arch_restriction = $2 ? 1 : 0;
+					my $addon_type = $field2addon_type{$field};
+					if (not defined($field)) {
+						warning("Cannot map ${field} to an add-on type (like \"both\", \"indep\" or \"arch\")");
+						error("Internal error: Cannot satisfy dh sequence add-on request for sequence ${sequence} via ${field}.");
+					}
+					if (defined($dh_bd_sequences{$sequence})) {
+						error("Saw $dep multiple times (last time in $field).  However dh only support that build-"
+							. 'dependency at most once across all Build-Depends(-Arch|-Indep) fields');
+					}
+					if ($has_profile_or_arch_restriction) {
+						require Dpkg::Deps;
+						my $dpkg_dep = Dpkg::Deps::deps_parse($dep, build_profiles => \@profiles, build_dep => 1,
+							reduce_restrictions => 1);
+						# If dpkg reduces it to nothing, then it was not relevant for us after all
+						next if not $dpkg_dep;
+					}
+					$dh_bd_sequences{$sequence} = $addon_type;
+				}
 			}
 		}
 		$compat_from_bd = $final_level // -1;
@@ -1598,50 +1838,97 @@ sub getpackages {
 		$compat_from_bd = -1;
 	}
 
+	%seen_fields = ();
+
 	while (<$fd>) {
 		chomp;
 		s/\s+$//;
-		if (/^Package:\s*+(.*)/i) {
-			$package=$1;
-			# Detect duplicate package names in the same control file.
-			if (! $seen{$package}) {
-				$seen{$package}=1;
-			}
-			else {
-				error("debian/control has a duplicate entry for $package");
-			}
-			$included_in_build_profile=1;
-		} elsif (/^Section:\s*+(.*)$/i) {
-			$section = $1;
-		} elsif (/^Architecture:\s*+(.*)/i) {
-			$arch=$1;
-		} elsif (/^(?:X[BC]*-)?Package-Type:\s*+(.*)/i) {
-			$package_type=$1;
-		} elsif (/^Multi-Arch:\s*(.*)/i) {
-			$multiarch = $1;
-		} elsif (/^X-DH-Build-For-Type:\s*+(.*)/i) {
-			$cross_type = $1;
-			if ($cross_type ne 'host' and $cross_type ne 'target') {
-				error("Unknown value of X-DH-Build-For-Type \"$cross_type\" at debian/control:$.");
-			}
-		} elsif (/^Build-Profiles:\s*+(.*)/i) {
-			# rely on libdpkg-perl providing the parsing functions
-			# because if we work on a package with a Build-Profiles
-			# field, then a high enough version of dpkg-dev is needed
-			# anyways
-			my $build_profiles=$1;
-			eval {
-				require Dpkg::BuildProfiles;
-				my @restrictions=Dpkg::BuildProfiles::parse_build_profiles($build_profiles);
-				if (@restrictions) {
-					$included_in_build_profile=Dpkg::BuildProfiles::evaluate_restriction_formula(\@restrictions, \@profiles);
-				}
-			};
-			if ($@) {
-				error("The control file has a Build-Profiles field. Requires libdpkg-perl >= 1.17.14");
-			}
+		if (m/^\#/) {
+			# Skip unless EOF for the special case where the last line
+			# is a comment line directly after the last stanza.  In
+			# that case we need to "commit" the last stanza as well or
+			# we end up omitting the last package.
+			next if not eof;
+			$_  = '';
 		}
 
+
+		if (/^\s/) {
+			# Continuation line
+			if (not %seen_fields) {
+				error("Continuation line seen outside stanza in debian/control (line $.)");
+			}
+		} elsif (not $_ and not %seen_fields) {
+			# Ignore empty lines before first stanza
+			next;
+		} elsif ($_) {
+			my ($field_name, $value);
+
+			if (m/^($DEB822_FIELD_REGEX):\s*(.*)/o) {
+				($field_name, $value) = (lc($1), $2);
+				if (exists($seen_fields{$field_name})) {
+					my $first_time = $seen_fields{$field_name};
+					error("${field_name}-field appears twice in the same stanza of debian/control. " .
+						  "First time on line $first_time, second time: $.");
+				}
+				$seen_fields{$field_name} = $.;
+				$bd_field_value = undef;
+			} else {
+				# Invalid file
+				error("Parse error in debian/control, line $., read: $_");
+			}
+
+			if ($field_name eq 'package') {
+				$package = $value;
+				# Detect duplicate package names in the same control file.
+				if (! $seen{$package}) {
+					$seen{$package}=1;
+				} else {
+					error("debian/control has a duplicate entry for $package");
+				}
+				if ($package !~ $valid_pkg_re) {
+					error('Package-field must be a valid package name, ' .
+						  "got: \"${package}\", should match \"${valid_pkg_re}\"");
+				}
+				$included_in_build_profile=1;
+			} elsif ($field_name eq 'section') {
+				$section = $value;
+			} elsif ($field_name eq 'architecture') {
+				$arch = $value;
+			} elsif ($field_name =~ m/^(?:x[bc]*-)?package-type$/) {
+				if (defined($package_type)) {
+					my $help = "(issue seen prior \"Package\"-field)";
+					$help = "for package ${package}" if $package;
+					error("Multiple definitions of (X-)Package-Type in line $. ${help}");
+				}
+				$package_type = $value;
+			} elsif ($field_name eq 'multi-arch') {
+				$multiarch = $value;
+			} elsif ($field_name eq 'x-dh-build-for-type') {
+				$cross_type = $value;
+				if ($cross_type ne 'host' and $cross_type ne 'target') {
+					error("Unknown value of X-DH-Build-For-Type \"$cross_type\" at debian/control:$.");
+				}
+			} elsif ($field_name eq 'build-profiles') {
+				# rely on libdpkg-perl providing the parsing functions
+				# because if we work on a package with a Build-Profiles
+				# field, then a high enough version of dpkg-dev is needed
+				# anyways
+				my $build_profiles = $value;
+				eval {
+					require Dpkg::BuildProfiles;
+					my @restrictions=Dpkg::BuildProfiles::parse_build_profiles($build_profiles);
+					if (@restrictions) {
+						$included_in_build_profile = Dpkg::BuildProfiles::evaluate_restriction_formula(
+							\@restrictions,
+							\@profiles);
+					}
+				};
+				if ($@) {
+					error("The control file has a Build-Profiles field. Requires libdpkg-perl >= 1.17.14");
+				}
+			}
+		}
 		if (!$_ or eof) { # end of stanza.
 			if ($package) {
 				$package_types{$package}=$package_type // 'deb';
@@ -1667,8 +1954,8 @@ sub getpackages {
 							$included = 1 if samearch($desired_arch, $arch);
 						}
 						if ($included) {
-							push(@{$packages_by_type{'arch'}}, $package);
-							push(@{$packages_by_type{'both'}}, $package);
+								push(@{$packages_by_type{'arch'}}, $package);
+								push(@{$packages_by_type{'both'}}, $package);
 						}
 					}
 				}
@@ -1678,11 +1965,10 @@ sub getpackages {
 			$cross_type = undef;
 			$arch='';
 			$section='';
+			%seen_fields = ();
 		}
 	}
 	close($fd);
-
-	return @{$packages_by_type{$type}};
 }
 
 # Return true if we should use root.
@@ -1829,6 +2115,14 @@ sub is_udeb {
 	}
 }
 
+# Only useful for dh(1)
+sub bd_dh_sequences {
+	# Use $sourcepackage as check because %dh_bd_sequence can be empty
+	# after running getpackages().
+	getpackages() if not defined($sourcepackage);
+	return \%dh_bd_sequences;
+}
+
 sub _concat_slurp_script_files {
 	my (@files) = @_;
 	my $res = '';
@@ -1841,43 +2135,67 @@ sub _concat_slurp_script_files {
 	return $res;
 }
 
+sub _substitution_generator {
+	my ($input) = @_;
+	my $cache = {};
+	return sub {
+		my ($orig_key) = @_;
+		return $cache->{$orig_key} if exists($cache->{$orig_key});
+		my $value = exists($input->{$orig_key}) ? $input->{$orig_key} : undef;
+		if (not defined($value)) {
+			if ($orig_key =~ m/^DEB_(?:BUILD|HOST|TARGET)_/) {
+				$value = dpkg_architecture_value($orig_key);
+			} elsif ($orig_key =~ m{^ENV[.](\S+)$}) {
+				$value = $ENV{$1} // '';
+			}
+		} elsif (ref($value) eq 'CODE') {
+			$value = $value->($orig_key);
+		} elsif ($value =~ s/^@//) {
+			$value = _concat_slurp_script_files($value);
+		}
+		$cache->{$orig_key} = $value;
+		return $value;
+	};
+}
+
 # Handles #DEBHELPER# substitution in a script; also can generate a new
 # script from scratch if none exists but there is a .debhelper file for it.
 sub debhelper_script_subst {
-	my $package=shift;
-	my $script=shift;
-	
+	my ($package, $script, $extra_vars) = @_;
+
 	my $tmp=tmpdir($package);
 	my $ext=pkgext($package);
 	my $file=pkgfile($package,$script);
+	my %variables = defined($extra_vars) ? %{$extra_vars} : ();
 	my $service_script = generated_file($package, "${script}.service", 0);
 	my @generated_scripts = ("debian/$ext$script.debhelper", $service_script);
+	my $subst;
+	@generated_scripts = grep { -f } @generated_scripts;
 	if ($script eq 'prerm' or $script eq 'postrm') {
 		@generated_scripts = reverse(@generated_scripts);
 	}
-	@generated_scripts = grep { -f } @generated_scripts;
+	if (not exists($variables{'DEBHELPER'})) {
+		$variables{'DEBHELPER'} = sub {
+			return _concat_slurp_script_files(@generated_scripts);
+		};
+	}
+	$subst = _substitution_generator(\%variables);
 
 	if ($file ne '') {
-		if (@generated_scripts) {
-			if ($dh{VERBOSE}) {
-				verbose_print('cp -f ' . escape_shell($file) . " $tmp/DEBIAN/$script");
-				verbose_print("perl -p -i -e \"s~#DEBHELPER#~qx{cat @generated_scripts}~eg\" $tmp/DEBIAN/$script");
+		if ($dh{VERBOSE}) {
+			verbose_print('cp -f ' . escape_shell($file) . " $tmp/DEBIAN/$script");
+			verbose_print("[META] Replace #TOKEN#s in \"$tmp/DEBIAN/$script\"");
+		}
+		if (not $dh{NO_ACT}) {
+			my $regex = qr{#(${MAINTSCRIPT_TOKEN_REGEX})#}o;
+			open(my $out_fd, '>', "$tmp/DEBIAN/$script") or error("open($tmp/DEBIAN/$script) failed: $!");
+			open(my $in_fd, '<', $file) or error("open($file) failed: $!");
+			while (my $line = <$in_fd>) {
+				$line =~ s{$regex}{$subst->($1) // "#${1}#"}ge;
+				print {$out_fd} $line;
 			}
-			# Add this into the script, where it has #DEBHELPER#
-			my $text = _concat_slurp_script_files(@generated_scripts);
-			if (not $dh{NO_ACT}) {
-				open(my $out_fd, '>', "$tmp/DEBIAN/$script") or error("open($tmp/DEBIAN/$script) failed: $!");
-				open(my $in_fd, '<', $file) or error("open($file) failed: $!");
-				while (my $line = <$in_fd>) {
-					$line =~ s/#DEBHELPER#/$text/g;
-					print {$out_fd} $line;
-				}
-				close($in_fd);
-				close($out_fd) or error("close($tmp/DEBIAN/$script) failed: $!");
-			}
-		} else {
-			# Just get rid of any #DEBHELPER# in the script.
-			doit({ stdout => "$tmp/DEBIAN/$script" }, 'sed', 's/#DEBHELPER#//', $file);
+			close($in_fd);
+			close($out_fd) or error("close($tmp/DEBIAN/$script) failed: $!");
 		}
 		reset_perm_and_owner('0755', "$tmp/DEBIAN/$script");
 	}
@@ -2100,6 +2418,42 @@ sub get_source_date_epoch {
 	return $ENV{SOURCE_DATE_EPOCH} = $timestamp->epoch();
 }
 
+# Setup the build ENV by setting dpkg-buildflags (via set_buildflags()) plus
+# cleaning up HOME (etc) in compat 13+
+sub setup_buildenv {
+	set_buildflags();
+	if (not compat(12)) {
+		setup_home_and_xdg_dirs();
+	}
+}
+
+sub setup_home_and_xdg_dirs {
+	my $home_dir = generated_file('_source', 'home', 0);
+	my $xdg_rundir = generated_file('_source', 'xdg-runtime-dir', 0);
+	my $creating_rundir = -d $xdg_rundir ? 0 : 1;
+	my @paths = (
+		$home_dir,
+		$xdg_rundir,
+	);
+	my @clear_env = qw(
+		XDG_CACHE_HOME
+		XDG_CONFIG_DIRS
+		XDG_CONFIG_HOME
+		XDG_DATA_HOME
+		XDG_DATA_DIRS
+	);
+	install_dir(@paths);
+	if ($creating_rundir) {
+		chmod(0700, $xdg_rundir) == 1 or warning("chmod(0700, \"$xdg_rundir\") failed: $! (ignoring)");
+	}
+	for my $envname (@clear_env) {
+		delete($ENV{$envname});
+	}
+	$ENV{'HOME'} = $home_dir;
+	$ENV{'XDG_RUNTIME_DIR'} = $xdg_rundir;
+	return;
+}
+
 # Sets environment variables from dpkg-buildflags. Avoids changing
 # any existing environment variables.
 sub set_buildflags {
@@ -2272,13 +2626,13 @@ sub open_gz {
 	};
 	if ($@) {
 		open($fd, '-|', 'gzip', '-dc', $file)
-		  or die("gzip -dc $file failed: $!");
+		  or error("gzip -dc $file failed: $!");
 	} else {
 		# Pass ":unix" as well due to https://rt.cpan.org/Public/Bug/Display.html?id=114557
 		# Alternatively, we could ensure we always use "POSIX::_exit".  Unfortunately,
 		# loading POSIX is insanely slow.
 		open($fd, '<:unix:gzip', $file)
-		  or die("open $file [<:unix:gzip] failed: $!");
+		  or error("open $file [<:unix:gzip] failed: $!");
 	}
 	return $fd;
 }
@@ -2358,6 +2712,15 @@ sub is_so_or_exec_elf_file {
 	return 1;
 }
 
+sub _has_shbang_line {
+	my ($file) = @_;
+	open(my $fd, '<', $file) or error("open $file: $!");
+	my $line = <$fd>;
+	close($fd);
+	return 1 if (defined($line) and substr($line, 0, 2) eq '#!');
+	return 0;
+}
+
 # Returns true iff the given argument is an empty directory.
 # Corner-cases:
 #  - false if not a directory
@@ -2491,17 +2854,54 @@ sub dbgsym_tmpdir {
 	return "debian/.debhelper/${package}/dbgsym-root";
 }
 
+sub perl_cross_incdir {
+	return if !is_cross_compiling();
+
+	# native builds don't currently need this so only load it on demand
+	require Config; Config->import();
+
+	my $triplet = dpkg_architecture_value("DEB_HOST_MULTIARCH");
+	my $perl_version = $Config::Config{version};
+	my $incdir = "/usr/lib/$triplet/perl/cross-config-${perl_version}";
+	return undef if !-e "$incdir/Config.pm";
+	return $incdir;
+}
 
 {
 	my %known_packages;
+	sub is_known_package {
+		my ($package) = @_;
+		%known_packages = map { $_ => 1 } getpackages() if not %known_packages;
+		return 1 if exists($known_packages{$package});
+		return 0
+	}
+
 	sub assert_opt_is_known_package {
 		my ($package, $method) = @_;
-		%known_packages = map { $_ => 1 } getpackages() if not %known_packages;
-		if (not exists($known_packages{$package})) {
+		if (not is_known_package($package)) {
 			error("Requested unknown package $package via $method, expected one of: " . join(' ', getpackages()));
 		}
 		return 1;
 	}
 }
 
+{
+	my $_disable_file_seccomp;
+	sub _internal_optional_file_args {
+		if (not defined($_disable_file_seccomp)) {
+			my $consider_disabling_seccomp = 0;
+			if ($ENV{'FAKEROOTKEY'} or ($ENV{'LD_PRELOAD'}//'') =~ m/fakeroot/) {
+				$consider_disabling_seccomp = 1;
+			}
+			if ($consider_disabling_seccomp) {
+				my $has_no_sandbox = (qx_cmd('file', '--help') // '') =~ m/--no-sandbox/;
+				$consider_disabling_seccomp = 0 if not $has_no_sandbox;
+			}
+			$_disable_file_seccomp = $consider_disabling_seccomp;
+		}
+		return ('--no-sandbox') if $_disable_file_seccomp;
+		return;
+	}
+}
+
 1
